"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = __importDefault(require("axios"));
const qrcode_data_pb_1 = require("../generated/qrcode_data_pb");
const common_1 = require("../utils/common");
const CryptoService_1 = __importDefault(require("./CryptoService"));
const index_1 = require("../index");
const EmailDataMessageV1 = qrcode_data_pb_1.vereign.protobuf.qrcode_data.EmailData_V1;
const KeyDataMessageV1 = qrcode_data_pb_1.vereign.protobuf.qrcode_data.KeyData_V1;
const WrapperDataMessage = qrcode_data_pb_1.vereign.protobuf.qrcode_data.WrapperData;
const EMAIL_DATA_V1 = "EmailData_V1";
const encodeEmailData = (emailData) => {
    const emailDataMessage = EmailDataMessageV1.fromObject(emailData);
    const emailDataBuffer = EmailDataMessageV1.encode(emailDataMessage).finish();
    const wrappedData = {
        version: 1,
        className: EMAIL_DATA_V1,
        data: emailDataBuffer,
    };
    const wrappedDataMessage = WrapperDataMessage.fromObject(wrappedData);
    const wrappedDataBuffer = WrapperDataMessage.encode(wrappedDataMessage).finish();
    return common_1.arrayBufferToBase64(wrappedDataBuffer);
};
const decodeEmailData = (binary) => {
    const wrappedDataMessage = WrapperDataMessage.decode(common_1.ensureUint8Array(binary));
    const wrappedData = WrapperDataMessage.toObject(wrappedDataMessage);
    if (wrappedData.className === EMAIL_DATA_V1) {
        const emailDataMessage = EmailDataMessageV1.decode(common_1.ensureUint8Array(wrappedData.data));
        const messageData = EmailDataMessageV1.toObject(emailDataMessage, {
            bytes: String,
            arrays: true,
            objects: true,
        });
        return messageData;
    }
    else {
        return null;
    }
};
const encodeKeyDataPair = (keyData) => {
    const keyDataMessage = KeyDataMessageV1.fromObject(keyData);
    const buffer = KeyDataMessageV1.encode(keyDataMessage).finish();
    return common_1.arrayBufferToBase64(buffer);
};
const decodeKeyDataPair = (binary) => {
    const keyDataPairMessage = KeyDataMessageV1.decode(common_1.ensureUint8Array(binary));
    return KeyDataMessageV1.toObject(keyDataPairMessage, {
        bytes: String,
    });
};
const breakQrCodeData = (data, headBytesSize = 32) => {
    if (data.byteLength <= headBytesSize) {
        return {
            head: data,
            tail: null,
        };
    }
    else {
        const dataUint8Array = new Uint8Array(data);
        return {
            head: dataUint8Array.slice(0, headBytesSize).buffer,
            tail: dataUint8Array.slice(headBytesSize, dataUint8Array.length).buffer,
        };
    }
};
const assembleQrCodeData = (head, tail) => {
    return Buffer.concat([Buffer.from(head), Buffer.from(tail)]);
};
const computeQrCodeHash = (emailData) => __awaiter(void 0, void 0, void 0, function* () {
    let attachments = [];
    if (emailData.attachments) {
        attachments = emailData.attachments.map((attachment) => {
            //TODO: check if we can use attachment hash, attachment hashAlg
            if (attachment.url) {
                return {
                    name: attachment.name,
                    size: attachment.size,
                    url: attachment.url,
                };
            }
            return {
                name: attachment.name,
                size: attachment.size,
            };
        });
    }
    const dataForHashing = {
        sender: emailData.sender,
        subject: emailData.subject,
        recipients: emailData.recipients,
        attachments,
    };
    const promises = Object.values(dataForHashing).map((value) => {
        const string = Buffer.from(JSON.stringify(value, (key, value) => value instanceof Object && !(value instanceof Array)
            ? Object.keys(value)
                .sort()
                .reduce((sorted, key) => {
                sorted[key] = value[key];
                return sorted;
            }, {})
            : value)).toString();
        return index_1.CryptoService.SHA256(string);
    });
    const hashArray = yield Promise.all(promises);
    const hashesAsAstring = hashArray
        .map(common_1.arrayBufferToBase64)
        .sort()
        .join("\n");
    const resultBuffer = yield index_1.CryptoService.SHA256(hashesAsAstring);
    return common_1.arrayBufferToBase64(resultBuffer);
});
const calculateQRCodeSignature = (accountPrivateKey, qrCodeHash) => __awaiter(void 0, void 0, void 0, function* () {
    const signature = yield index_1.CryptoService.signRSA(accountPrivateKey, Buffer.from(qrCodeHash));
    return common_1.arrayBufferToBase64(signature);
});
const verifyQrCodeSignature = (publicKey, qrCodeSignature, recipientQRCodeHash) => __awaiter(void 0, void 0, void 0, function* () {
    const qrCodeSignatureAsBuffer = common_1.base64ToArrayBuffer(qrCodeSignature);
    const recipientQRCodeHashAsBuffer = Buffer.from(recipientQRCodeHash);
    const result = yield index_1.CryptoService.verifyRSASignature(publicKey, recipientQRCodeHashAsBuffer, qrCodeSignatureAsBuffer);
    return result;
});
/**
 * Higher order function to perform complex operations with the QR code data
 * @param cloudFlareServiceInstance = new CloudflareService("CDN_RUL", "DEFAULT_BUCKET")
 * @param landingPageServiceUrl - URL of the service responsible for RSA decryption
 * @param cryptoService - instance of ICryptoService. Override only needed cryptographic functions.
 * The rest will be filled with default ones from CryptoServiceWeb/CryptoServiceNode
 */
const withServices = (cloudFlareServiceInstance, landingPageServiceUrl, cryptoService) => {
    if (cryptoService) {
        cryptoService = Object.assign(Object.assign({}, CryptoService_1.default), cryptoService);
    }
    else {
        cryptoService = CryptoService_1.default;
    }
    return {
        getMessageDataFromBase64: (base64) => __awaiter(void 0, void 0, void 0, function* () {
            const qrCodeDataPart = decodeKeyDataPair(base64);
            const base64SHA256 = yield cryptoService.SHA256(base64);
            const { data: backblazeDataPart, } = yield cloudFlareServiceInstance.fetchFile(`qrcode-${common_1.arrayBufferToHex(base64SHA256)}`);
            const assembledData = assembleQrCodeData(common_1.base64ToArrayBuffer(qrCodeDataPart.data), common_1.base64ToArrayBuffer(backblazeDataPart.qr_code_data));
            const { key: aesEncryptedSessionKey, data: storageIv, } = decodeKeyDataPair(backblazeDataPart.session_key);
            const rsaEncryptedSessionKey = yield cryptoService.decryptAESGCM(common_1.base64ToArrayBuffer(aesEncryptedSessionKey), common_1.base64ToArrayBuffer(qrCodeDataPart.key), common_1.base64ToArrayBuffer(storageIv));
            const { data: { key: encodedSessionKey }, } = yield axios_1.default({
                url: `${landingPageServiceUrl}/api/hsm/decrypt`,
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                data: {
                    key_signature: backblazeDataPart.key_signature,
                    encrypted_key: rsaEncryptedSessionKey,
                },
            });
            const decodedSessionKey = decodeKeyDataPair(encodedSessionKey);
            const decryptedEmailData = yield cryptoService.decryptAESGCM(assembledData, common_1.base64ToArrayBuffer(decodedSessionKey.key), common_1.base64ToArrayBuffer(decodedSessionKey.data));
            const decompressedEmailData = common_1.decompressData(decryptedEmailData);
            return decodeEmailData(decompressedEmailData);
        }),
    };
};
exports.default = {
    encodeEmailData,
    decodeEmailData,
    encodeKeyDataPair,
    decodeKeyDataPair,
    breakQrCodeData,
    assembleQrCodeData,
    computeQrCodeHash,
    calculateQRCodeSignature,
    verifyQrCodeSignature,
    withServices,
};
