import axios from "axios";
import { StatusData } from "../types";
import StatusesService from "./StatusesService";
import { base64ToArrayBuffer } from "../utils/common";

const STATUS_FILE_PREFIX = "bs";

interface RawStatusData {
  Status: string;
  BatchID: string;
  TransactionID: string;
}

class CloudflareService {
  private _cdnUrl = "";
  private _defaultBucket = "";
  private _statusesService;

  constructor(cdnUrl: string, defaultBucket: string) {
    if (!cdnUrl) {
      throw new Error("CDN URL Must not be empty");
    }

    if (!defaultBucket) {
      throw new Error("Default bucket must not be empty");
    }

    this._cdnUrl = cdnUrl;
    this._defaultBucket = defaultBucket;
    this._statusesService = new StatusesService();
  }

  async fetchFile<T>(
    fileName: string,
    bucket = this._defaultBucket
  ): Promise<{ data: T; uploadTime: number }> {
    // v?= is a dirty fix to prevent caching of the CORS error response in case file is missing.
    // Because of browser caching CORS error in case of missing file, it's not possible to retrieve it when it's being uploaded.
    // Dig into https://www.backblaze.com/b2/docs/cors_rules.html and fix
    const response = await axios(
      `${this._cdnUrl}/file/${bucket}/${fileName}?v=${Math.random()}`
    );

    const data = response.data as T;

    const uploadTime = response.headers["x-bz-upload-timestamp"]
      ? parseInt(response.headers["x-bz-upload-timestamp"])
      : -1;

    return { data, uploadTime };
  }

  async fetchStatusData(statusFileName: string): Promise<StatusData> {
    const {
      data: rawStatusData,
      uploadTime,
    } = await this.fetchFile<RawStatusData>(statusFileName);
    let decodedResult;
    try {
      decodedResult = this._statusesService.decodeStatusObject(
        new Uint8Array(base64ToArrayBuffer(rawStatusData.Status))
      );
    } catch (e) {
      throw new Error("Error decoding status object");
    }

    return {
      status: decodedResult.statusObject,
      batchId: rawStatusData.BatchID,
      transactionHash: rawStatusData.TransactionID,
      statusRaw: rawStatusData.Status,
      statusClassName: decodedResult.className,
      timestamp: uploadTime,
    };
  }

  async readStatuses(statusId: string): Promise<Array<StatusData>> {
    let finished = false;
    let currentIndex = 1;

    const statuses = [];
    while (!finished) {
      try {
        const statusData = await this.fetchStatusData(
          `${STATUS_FILE_PREFIX}_${statusId}_${currentIndex}`
        );
        statuses.push(statusData);
        currentIndex++;
      } catch (error) {
        // Axios error. Finishing fetching statuses.
        if (error.isAxiosError) {
          finished = true;
        } else {
          throw error;
        }
      }
    }

    return statuses;
  }

  async getMerkleTree(fileName: string): Promise<Array<string>> {
    const { data: merkleTreeFileContents } = await this.fetchFile<string[]>(
      fileName
    );

    const statusesData: Array<StatusData> = await Promise.all(
      merkleTreeFileContents.map((statusFileName) => {
        return this.fetchStatusData(statusFileName);
      })
    );

    const statusesMap: {
      [key: string]: StatusData;
    } = statusesData.reduce((acc, statusData: StatusData, index) => {
      Object.assign(acc, {
        [merkleTreeFileContents[index].toLowerCase()]: statusData,
      });

      return acc;
    }, {});

    return Object.keys(statusesMap)
      .sort()
      .map((key) => statusesMap[key].statusRaw);
  }
}

export default CloudflareService;
