import { Zlib } from "zlibjs/bin/zlib.min";
import { CryptoService } from "../index";

export const arrayBufferToBase64 = (buffer: ArrayBuffer): string => {
  return Buffer.from(buffer).toString("base64");
};

export const base64ToArrayBuffer = (base64: string): ArrayBuffer => {
  return Buffer.from(base64, "base64");
};

export const ensureBase64 = (data: string | ArrayBuffer): string => {
  return typeof data === "string" ? data : arrayBufferToBase64(data);
};

export const ensureArrayBuffer = (data: string | ArrayBuffer): ArrayBuffer => {
  return typeof data === "string" ? base64ToArrayBuffer(data) : data;
};

export const ensureUint8Array = (
  data: string | ArrayBuffer | Uint8Array
): Uint8Array => {
  return data instanceof Uint8Array
    ? data
    : typeof data === "string"
    ? new Uint8Array(base64ToArrayBuffer(data))
    : data instanceof ArrayBuffer
    ? new Uint8Array(data)
    : data;
};

export const compressData = (binary: string | ArrayBuffer): ArrayBuffer => {
  const deflate = new Zlib.Deflate(ensureUint8Array(binary));
  return deflate.compress();
};

export const decompressData = (binary: string | ArrayBuffer): ArrayBuffer => {
  const inflate = new Zlib.Inflate(ensureUint8Array(binary));
  return inflate.decompress();
};

export const arrayBufferToHex = (buffer: ArrayBuffer): string => {
  return [...new Uint8Array(buffer)]
    .map((b) => b.toString(16).padStart(2, "0"))
    .join("");
};

export const encodeBase64Url = (base64: string): string => {
  return base64.replace(/\+/g, "-").replace(/\//g, "_");
};

export const decodeBase64URL = (input: string): string => {
  return input.replace(/_/g, "/").replace(/-/g, "+");
};

export const getMerkleTreeRootHash = async (
  leaves: Buffer[]
): Promise<Buffer> => {
  let hashes = [...leaves];
  while (hashes.length != 1) {
    let pair: Buffer;

    const newHashes = [];
    for (let i = 0; i < hashes.length; i++) {
      const hash = hashes[i];
      pair = pair ? Buffer.concat([pair, hash]) : hash;

      if (i % 2) {
        newHashes.push(await CryptoService.SHA256(pair));
        pair = null;
      } else {
        // If array has odd items, transfer the last one to the upper level of the tree
        if (i === hashes.length - 1) {
          newHashes.push(hash);
        }
      }
    }

    hashes = newHashes;
  }

  return hashes[0];
};

export default {
  compressData,
  decompressData,
  arrayBufferToHex,
  arrayBufferToBase64,
  base64ToArrayBuffer,
  ensureUint8Array,
  ensureArrayBuffer,
  ensureBase64,
};
